@@ -1,20 +1,27 @@
-from typing import Any, Tuple, Union
+from itertools import chain
+from typing import Any, Dict, Tuple, Union
 
-import geopandas as gpd
 import numpy as np
-import pyproj
-import shapely.geometry as sg
+import pandas as pd
 import xarray as xr
 from numba_celltree import CellTree2d
 from scipy.sparse import coo_matrix, csr_matrix
 from scipy.sparse.csgraph import reverse_cuthill_mckee
+from xarray.core.utils import either_dict_or_kwargs
 
-from .. import connectivity, conversion
-from .. import meshkernel_utils as mku
-from ..typing import BoolArray, FloatArray, FloatDType, IntArray, IntDType, SparseMatrix
-from ..voronoi import voronoi_topology
-from . import ugrid_io
-from .ugridbase import AbstractUgrid
+from xugrid import conversion
+from xugrid import meshkernel_utils as mku
+from xugrid.constants import (
+    BoolArray,
+    FloatArray,
+    FloatDType,
+    IntArray,
+    IntDType,
+    SparseMatrix,
+)
+from xugrid.ugrid import connectivity, conventions
+from xugrid.ugrid.ugridbase import AbstractUgrid, as_pandas_index
+from xugrid.ugrid.voronoi import voronoi_topology
 
 
 def section_coordinates(
@@ -49,13 +56,25 @@ class Ugrid2d(AbstractUgrid):
     node_y: ndarray of floats
     fill_value: int
     face_node_connectivity: ndarray of integers
+    name: string, optional
+        Mesh name. Defaults to "mesh2d".
     edge_node_connectivity: ndarray of integers, optional
     dataset: xr.Dataset, optional
-    name: string, optional
+    indexes: Dict[str, str], optional
+        When a dataset is provided, a mapping from the UGRID role to the dataset
+        variable name. E.g. {"face_x": "mesh2d_face_lon"}.
+    projected: bool, optional
+        Whether node_x and node_y are longitude and latitude or projected x and
+        y coordinates. Used to write the appropriate standard_name in the
+        coordinate attributes.
     crs: Any, optional
         Coordinate Reference System of the geometry objects. Can be anything accepted by
         :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
         such as an authority string (eg "EPSG:4326") or a WKT string.
+    attrs: Dict[str, str], optional
+        UGRID topology attributes. Should not be provided together with
+        dataset: if other names are required, update the dataset instead.
+        A name entry is ignored, as name is given explicitly.
     """
 
     def __init__(
@@ -64,13 +83,19 @@ class Ugrid2d(AbstractUgrid):
         node_y: FloatArray,
         fill_value: int,
         face_node_connectivity: Union[IntArray, SparseMatrix],
+        name: str = "mesh2d",
         edge_node_connectivity: IntArray = None,
         dataset: xr.Dataset = None,
-        name: str = None,
+        indexes: Dict[str, str] = None,
+        projected: bool = True,
         crs: Any = None,
+        attrs: Dict[str, str] = None,
     ):
         self.node_x = np.ascontiguousarray(node_x)
         self.node_y = np.ascontiguousarray(node_y)
+        self.fill_value = fill_value
+        self.name = name
+        self.projected = projected
 
         if isinstance(face_node_connectivity, np.ndarray):
             face_node_connectivity = face_node_connectivity
@@ -83,14 +108,12 @@ class Ugrid2d(AbstractUgrid):
                 "face_node_connectivity should be an array of integers or a sparse matrix"
             )
 
-        self.fill_value = fill_value
         self.face_node_connectivity = connectivity.counterclockwise(
             face_node_connectivity, self.fill_value, self.node_coordinates
         )
-        if name is None:
-            name = ugrid_io.UGRID2D_DEFAULT_NAME
-        self.name = name
-        self.topology_attrs = None
+
+        self._initialize_indexes_attrs(name, dataset, indexes, attrs)
+        self._dataset = dataset
 
         # Optional attributes, deferred initialization
         # Meshkernel
@@ -98,6 +121,8 @@ class Ugrid2d(AbstractUgrid):
         self._meshkernel = None
         # Celltree
         self._celltree = None
+        # Area
+        self._area = None
         # Centroids
         self._centroids = None
         # Bounds
@@ -123,12 +148,9 @@ class Ugrid2d(AbstractUgrid):
         if crs is None:
             self.crs = None
         else:
+            import pyproj
+
             self.crs = pyproj.CRS.from_user_input(crs)
-        # Store dataset
-        if dataset is None:
-            dataset = self.topology_dataset()
-        self.dataset = dataset
-        self.topology_attrs = dataset[self.name].attrs
 
     def _clear_geometry_properties(self):
         """Clear all properties that may have been invalidated"""
@@ -137,6 +159,8 @@ class Ugrid2d(AbstractUgrid):
         self._meshkernel = None
         # Celltree
         self._celltree = None
+        # Area
+        self._area = None
         # Centroids
         self._centroids = None
         # Bounds
@@ -152,8 +176,53 @@ class Ugrid2d(AbstractUgrid):
         self._voronoi_topology = None
         self._centroid_triangulation = None
 
-    @staticmethod
-    def from_dataset(dataset: xr.Dataset):
+    @classmethod
+    def from_meshkernel(
+        cls,
+        mesh,
+        name: str = "mesh2d",
+        projected: bool = True,
+        crs: Any = None,
+    ):
+        """
+        Create a 2D UGRID topology from a MeshKernel Mesh2d object.
+
+        Parameters
+        ----------
+        mesh: MeshKernel.Mesh2d
+        name: str
+            Mesh name. Defaults to "mesh2d".
+        projected: bool
+            Whether node_x and node_y are longitude and latitude or projected x and
+            y coordinates. Used to write the appropriate standard_name in the
+            coordinate attributes.
+        crs: Any, optional
+            Coordinate Reference System of the geometry objects. Can be anything accepted by
+            :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
+            such as an authority string (eg "EPSG:4326") or a WKT string.
+
+        Returns
+        -------
+        grid: Ugrid2d
+        """
+        n_face = len(mesh.nodes_per_face)
+        n_max_node = mesh.nodes_per_face.max()
+        fill_value = -1
+        face_node_connectivity = np.full((n_face, n_max_node), fill_value)
+        isnode = connectivity.ragged_index(n_face, n_max_node, mesh.nodes_per_face)
+        face_node_connectivity[isnode] = mesh.face_nodes
+        return cls(
+            mesh.node_x,
+            mesh.node_y,
+            fill_value=fill_value,
+            face_node_connectivity=face_node_connectivity,
+            name=name,
+            projected=projected,
+            crs=crs,
+        )
+
+    @classmethod
+    def from_dataset(cls, dataset: xr.Dataset, topology: str = None):
         """
         Extract the 2D UGRID topology information from an xarray Dataset.
 
@@ -164,7 +233,7 @@ class Ugrid2d(AbstractUgrid):
 
         Returns
         -------
-        grid: Ugrid2d
+        grid: Ugrid1dAdapter
         """
         ds = dataset
         if not isinstance(ds, xr.Dataset):
@@ -172,149 +241,136 @@ class Ugrid2d(AbstractUgrid):
                 "Ugrid should be initialized with xarray.Dataset. "
                 f"Received instead: {type(ds)}"
             )
+        if topology is None:
+            topology = cls._single_topology(ds)
+
+        indexes = {}
 
         # Collect names
-        mesh_topology = ugrid_io.get_topology_variable(ds)
-        ugrid_roles = ugrid_io.get_ugrid2d_variables(dataset, mesh_topology)
-        # Rename the arrays to their standard UGRID roles
-        topology_ds = ds[set(ugrid_roles.keys())]
-        ugrid_ds = topology_ds.rename(ugrid_roles)
-        # Coerce type and fill value
-        ugrid_ds["node_x"] = ugrid_ds["node_x"].astype(FloatDType)
-        ugrid_ds["node_y"] = ugrid_ds["node_y"].astype(FloatDType)
-        fill_value = -1
-        ugrid_ds["face_node_connectivity"] = ugrid_io.cast(
-            ugrid_ds["face_node_connectivity"], fill_value, IntDType
+        connectivity = ds.ugrid_roles.connectivity[topology]
+        coordinates = ds.ugrid_roles.coordinates[topology]
+        ugrid_vars = (
+            [topology]
+            + list(connectivity.values())
+            + list(chain.from_iterable(chain.from_iterable(coordinates.values())))
         )
-        if "edge_node_connectivity" in ugrid_ds:
-            edge_node_connectivity = (
-                ugrid_ds["edge_node_connectivity"].astype(IntDType).values
-            )
+        fill_value = -1
+
+        x_index = coordinates["node_coordinates"][0][0]
+        y_index = coordinates["node_coordinates"][1][0]
+        node_x_coordinates = ds[x_index].astype(FloatDType).values
+        node_y_coordinates = ds[y_index].astype(FloatDType).values
+
+        face_nodes = connectivity["face_node_connectivity"]
+        face_node_connectivity = cls._prepare_connectivity(
+            ds[face_nodes], fill_value, dtype=IntDType
+        ).values
+
+        edge_nodes = connectivity.get("edge_node_connectivity")
+        if edge_nodes:
+            edge_node_connectivity = cls._prepare_connectivity(
+                ds[edge_nodes], fill_value, dtype=IntDType
+            ).values
         else:
             edge_node_connectivity = None
 
-        return Ugrid2d(
-            ugrid_ds["node_x"].values,
-            ugrid_ds["node_y"].values,
+        indexes["node_x"] = x_index
+        indexes["node_y"] = y_index
+        projected = False  # TODO
+
+        return cls(
+            node_x_coordinates,
+            node_y_coordinates,
             fill_value,
-            ugrid_ds["face_node_connectivity"].values,
-            edge_node_connectivity,
-            topology_ds,
-            mesh_topology.name,
+            face_node_connectivity,
+            name=topology,
+            edge_node_connectivity=edge_node_connectivity,
+            dataset=ds[ugrid_vars],
+            indexes=indexes,
+            projected=projected,
+            crs=None,
         )
 
-    def topology_dataset(self):
-        """
-        Store the UGRID topology information in an xarray Dataset according to
-        the UGRID conventions.
+    def to_dataset(self, other: xr.Dataset = None) -> xr.Dataset:
+        node_x = self._indexes["node_x"]
+        node_y = self._indexes["node_y"]
+        face_nodes = self._attrs["face_node_connectivity"]
+        face_nodes_attrs = conventions.DEFAULT_ATTRS["face_node_connectivity"]
+        nmax_node_dim = self._attrs["max_face_nodes_dimension"]
+        edge_nodes = self._attrs["edge_node_connectivity"]
+        edge_nodes_attrs = conventions.DEFAULT_ATTRS["edge_node_connectivity"]
 
-        Returns
-        -------
-        ugrid_topology: xr.Dataset
-        """
-        return ugrid_io.ugrid2d_dataset(
-            self.node_x,
-            self.node_y,
-            self.fill_value,
-            self.face_node_connectivity,
-            self._edge_node_connectivity,
-            self.name,
-            self.topology_attrs,
-        )
+        data_vars = {
+            self.name: 0,
+            face_nodes: xr.DataArray(
+                data=self.face_node_connectivity,
+                attrs=face_nodes_attrs,
+                dims=(self.face_dimension, nmax_node_dim),
+            ),
+        }
+        if self.edge_node_connectivity is not None:
+            data_vars[edge_nodes] = xr.DataArray(
+                data=self.edge_node_connectivity,
+                attrs=edge_nodes_attrs,
+                dims=(self.edge_dimension, "two"),
+            )
 
-    def remove_topology(
-        self, obj: Union[xr.DataArray, xr.Dataset]
-    ) -> Union[xr.DataArray, xr.Dataset]:
-        """
-        Remove UGRID specific variables from the object.
+        dataset = xr.Dataset(data_vars, attrs={"Conventions": "CF-1.8 UGRID-1.0"})
 
-        Parameters
-        ----------
-        obj: Union[xr.DataArray, xr.Dataset]
+        if self._dataset:
+            dataset.update(self._dataset)
+        if other is not None:
+            dataset = dataset.merge(other)
+        if node_x not in dataset or node_y not in dataset:
+            dataset = self.assign_node_coords(dataset)
 
-        Returns
-        -------
-        sanitized: Union[xr.DataArray, xr.Dataset]
-        """
-        return self._remove_topology(obj, ugrid_io.UGRID2D_TOPOLOGY_VARIABLES)
-
-    def topology_coords(self, obj: Union[xr.DataArray, xr.Dataset]) -> dict:
-        """
-        Return a dictionary with the coordinates required for the dimension of
-        the object.
-
-        Parameters
-        ----------
-        obj: Union[xr.DataArray, xr.Dataset]
-
-        Returns
-        -------
-        coords: dict
-        """
-        coords = {}
-        dims = obj.dims
-        facedim = self.face_dimension
-        edgedim = self.edge_dimension
-        nodedim = self.node_dimension
-        attrs = self.topology_attrs
-
-        if facedim in dims:
-            # It's not a required attribute.
-            if "face_coordinates" in attrs:
-                name_x, name_y = attrs["face_coordinates"].split()
-            else:
-                name_x = f"{self.name}_face_x"
-                name_y = f"{self.name}_face_y"
-            coords[name_x] = (facedim, self.face_x)
-            coords[name_y] = (facedim, self.face_y)
-            attrs["face-coordinates"] = f"{name_x} {name_y}"
-
-        if edgedim in dims:
-            # It's not a required attribute.
-            if "edge_coordinates" in attrs:
-                name_x, name_y = attrs["edge_coordinates"].split()
-            else:
-                name_x = f"{self.name}_edge_x"
-                name_y = f"{self.name}_edge_y"
-            coords[name_x] = (edgedim, self.edge_x)
-            coords[name_y] = (edgedim, self.edge_y)
-            attrs["edge_coordinates"] = f"{name_x} {name_y}"
-
-        if nodedim in dims:
-            name_x, name_y = attrs["node_coordinates"].split()
-            coords[name_x] = (nodedim, self.node_x)
-            coords[name_y] = (nodedim, self.node_y)
-
-        return coords
+        dataset[self.name].attrs = self._filtered_attrs(dataset)
+        return dataset
 
     # These are all optional/derived UGRID attributes. They are not computed by
     # default, only when called upon.
     @property
-    def n_face(self):
+    def n_face(self) -> int:
         """
         Return the number of faces in the UGRID2D topology.
         """
         return self.face_node_connectivity.shape[0]
 
+    @property
+    def n_max_node_per_face(self) -> int:
+        """
+        Return the maximum number of nodes that a face can contain in the
+        UGRID2D topology.
+        """
+        return self.face_node_connectivity.shape[1]
+
+    @property
+    def n_node_per_face(self) -> IntArray:
+        return (self.face_node_connectivity != self.fill_value).sum(axis=1)
+
+    @property
+    def core_dimension(self):
+        return self.face_dimension
+
+    @property
+    def dimensions(self):
+        return {
+            self.node_dimension: self.n_node,
+            self.edge_dimension: self.n_edge,
+            self.face_dimension: self.n_face,
+        }
+
     @property
     def topology_dimension(self):
         """Highest dimensionality of the geometric elements: 2"""
         return 2
 
-    def _get_dimension(self, dim):
-        key = f"{dim}_dimension"
-        if key not in self.topology_attrs:
-            self.topology_attrs[key] = ugrid_io.UGrid.mesh2d_get_attributes(self.name)[
-                key
-            ]
-        return self.topology_attrs[key]
-
     @property
     def face_dimension(self):
         """
         Return the name of the face dimension.
         """
-        return self._get_dimension("face")
+        return self._attrs["face_dimension"]
 
     def _edge_connectivity(self):
         (
@@ -323,6 +379,7 @@ class Ugrid2d(AbstractUgrid):
         ) = connectivity.edge_connectivity(
             self.face_node_connectivity,
             self.fill_value,
+            self._edge_node_connectivity,
         )
 
     @property
@@ -350,6 +407,11 @@ class Ugrid2d(AbstractUgrid):
         """
         if self._face_edge_connectivity is None:
             self._edge_connectivity()
+            # if self._edge_node_connectivity is None:
+            #    self._edge_connectivity()
+            # else:
+            #    raise NotImplementedError
+        #                self._face_edge_connectivity =
         return self._face_edge_connectivity
 
     @property
@@ -370,6 +432,41 @@ class Ugrid2d(AbstractUgrid):
             )
         return self._centroids
 
+    @property
+    def area(self) -> FloatArray:
+        if self._area is None:
+            self._area = connectivity.area(
+                self.face_node_connectivity,
+                self.fill_value,
+                self.node_x,
+                self.node_y,
+            )
+        return self._area
+
+    @property
+    def face_bounds(self):
+        """
+        Returns a numpy array with columns ``minx, miny, maxx, maxy``,
+        describing the bounds of every face in the grid.
+
+        Returns
+        -------
+        face_bounds: np.ndarray of shape (n_face, 4)
+        """
+        x = self.node_x[self.face_node_connectivity]
+        y = self.node_y[self.face_node_connectivity]
+        isfill = self.face_node_connectivity == self.fill_value
+        x[isfill] = np.nan
+        y[isfill] = np.nan
+        return np.column_stack(
+            [
+                np.nanmin(x, axis=1),
+                np.nanmin(y, axis=1),
+                np.nanmax(x, axis=1),
+                np.nanmax(y, axis=1),
+            ]
+        )
+
     @property
     def face_x(self):
         """x-coordinate of centroid of every face"""
@@ -498,6 +595,11 @@ class Ugrid2d(AbstractUgrid):
                 self.node_face_connectivity,
                 self.node_coordinates,
                 self.centroids,
+                self.edge_face_connectivity,
+                self.edge_node_connectivity,
+                self.fill_value,
+                add_exterior=True,
+                add_vertices=False,
             )
             self._voronoi_topology = vertices, faces, face_index
         return self._voronoi_topology
@@ -582,6 +684,42 @@ class Ugrid2d(AbstractUgrid):
             )
         return self._celltree
 
+    def assign_face_coords(
+        self,
+        obj: Union[xr.DataArray, xr.Dataset],
+    ) -> Union[xr.DataArray, xr.Dataset]:
+        """
+        Assign face coordinates from the grid to the object.
+
+        Returns a new object with all the original data in addition to the new
+        node coordinates of the grid.
+
+        Parameters
+        ----------
+        obj: xr.DataArray or xr.Dataset
+
+        Returns
+        -------
+        assigned (same type as obj)
+        """
+        xname = self._indexes.get("face_x", f"{self.name}_face_x")
+        yname = self._indexes.get("face_y", f"{self.name}_face_y")
+        x_attrs = conventions.DEFAULT_ATTRS["face_x"][self.projected]
+        y_attrs = conventions.DEFAULT_ATTRS["face_y"][self.projected]
+        coords = {
+            xname: xr.DataArray(
+                data=self.face_x,
+                dims=(self.face_dimension,),
+                attrs=x_attrs,
+            ),
+            yname: xr.DataArray(
+                data=self.face_y,
+                dims=(self.face_dimension,),
+                attrs=y_attrs,
+            ),
+        }
+        return obj.assign_coords(coords)
+
     def locate_points(self, points: FloatArray):
         """
         Find in which face points are located.
@@ -641,6 +779,24 @@ class Ugrid2d(AbstractUgrid):
             & (self.face_y < ymax)
         )[0]
 
+    def compute_barycentric_weights(
+        self, points: FloatArray
+    ) -> Tuple[IntArray, FloatArray]:
+        """
+        Find in which face the points are located, and compute the barycentric
+        weight for every vertex of the face.
+
+        Parameters
+        ----------
+        points: ndarray of floats with shape ``(n_point, 2)``
+
+        Returns
+        -------
+        face_index: ndarray of integers with shape ``(n_points,)``
+        weights: ndarray of floats with shape ```(n_points, n_max_node)``
+        """
+        return self.celltree.compute_barycentric_weights(points)
+
     def rasterize_like(
         self, x: FloatArray, y: FloatArray
     ) -> Tuple[FloatArray, FloatArray, IntArray]:
@@ -691,6 +847,154 @@ class Ugrid2d(AbstractUgrid):
         y = np.arange(ymax - 0.5 * d, ymin, -d)
         return self.rasterize_like(x, y)
 
+    def topology_subset(
+        self, face_index: Union[BoolArray, IntArray], return_index: bool = False
+    ):
+        """
+        Create a new UGRID1D topology for a subset of this topology.
+
+        Parameters
+        ----------
+        face_index: 1d array of integers or bool
+            Edges of the subset.
+        return_index: bool, optional
+            Whether to return node_index, edge_index, face_index.
+
+        Returns
+        -------
+        subset: Ugrid2d
+        indexes: dict
+            Dictionary with keys node dimension, edge dimension, face dimension
+            and values their respective index. Only returned if return_index is
+            True.
+        """
+        if not isinstance(face_index, pd.Index):
+            face_index = as_pandas_index(face_index, self.n_face)
+
+        # The pandas index may only contain uniques. So if size matches, it's the identity.
+        if face_index.size == self.n_face:
+            # TODO: return self.copy instead?
+            if return_index:
+                indexes = {
+                    self.node_dimension: pd.RangeIndex(0, self.n_node),
+                    self.edge_dimension: pd.RangeIndex(0, self.n_edge),
+                    self.face_dimension: pd.RangeIndex(0, self.n_face),
+                }
+                return self, indexes
+            else:
+                return self
+
+        index = face_index.values
+        face_subset = self.face_node_connectivity[index]
+        node_index = np.unique(face_subset.ravel())
+        node_index = node_index[node_index != self.fill_value]
+        new_faces = connectivity.renumber(face_subset, self.fill_value)
+        node_x = self.node_x[node_index]
+        node_y = self.node_y[node_index]
+
+        edge_index = None
+        new_edges = None
+        if self.edge_node_connectivity is not None:
+            edge_index = np.unique(self.face_edge_connectivity[index].ravel())
+            edge_index = edge_index[edge_index != self.fill_value]
+            edge_subset = self.edge_node_connectivity[edge_index]
+            new_edges = connectivity.renumber(edge_subset)
+
+        grid = self.__class__(
+            node_x,
+            node_y,
+            self.fill_value,
+            new_faces,
+            name=self.name,
+            edge_node_connectivity=new_edges,
+            projected=self.projected,
+            crs=self.crs,
+            attrs=self._attrs,
+        )
+        if return_index:
+            indexes = {
+                self.node_dimension: pd.Index(node_index),
+                self.face_dimension: face_index,
+            }
+            if edge_index is not None:
+                indexes[self.edge_dimension] = pd.Index(edge_index)
+            return grid, indexes
+        else:
+            return grid
+
+    def clip_box(
+        self,
+        xmin: float,
+        ymin: float,
+        xmax: float,
+        ymax: float,
+    ):
+        xmin, ymin, xmax, ymax = self.bounds
+        bounds = [xmin, ymin, xmax, ymax]
+        face_index = self.locate_bounding_box(*bounds)
+        return self.topology_subset(face_index)
+
+    def isel(self, indexers=None, return_index=False, **indexers_kwargs):
+        """
+        Select based on node, edge, or face.
+
+        Face selection always results in a valid UGRID topology.
+        Node or edge selection may result in invalid topologies (incomplete
+        faces), and will error in such a case.
+
+        Parameters
+        ----------
+        indexers: dict of str to np.ndarray of integers or bools
+        return_index: bool, optional
+            Whether to return node_index, edge_index, face_index.
+
+        Returns
+        -------
+        obj: xr.Dataset or xr.DataArray
+        grid: Ugrid2d
+        indexes: dict
+            Dictionary with keys node dimension, edge dimension, face dimension
+            and values their respective index. Only returned if return_index is
+            True.
+        """
+        indexers = either_dict_or_kwargs(indexers, indexers_kwargs, "isel")
+        alldims = set(self.dimensions)
+        invalid = indexers.keys() - alldims
+        if invalid:
+            raise ValueError(
+                f"Dimensions {invalid} do not exist. Expected one of {alldims}"
+            )
+
+        indexers = {
+            k: as_pandas_index(v, self.dimensions[k]) for k, v in indexers.items()
+        }
+        nodedim, edgedim, facedim = self.dimensions
+        face_index = {}
+        if nodedim in indexers:
+            node_index = indexers[nodedim]
+            face_index[nodedim] = np.unique(
+                self.node_face_connectivity[node_index].data
+            )
+        if edgedim in indexers:
+            edge_index = indexers[edgedim]
+            index = np.unique(self.edge_face_connectivity[edge_index])
+            face_index[edgedim] = index[index != self.fill_value]
+        if facedim in indexers:
+            face_index[facedim] = indexers[facedim]
+
+        # Convert all to pandas index.
+        face_index = {k: as_pandas_index(v, self.n_face) for k, v in face_index.items()}
+
+        # Check the indexes against each other.
+        index = self._precheck(face_index)
+        grid, finalized_indexers = self.topology_subset(index, return_index=True)
+        self._postcheck(indexers, finalized_indexers)
+
+        if return_index:
+            return grid, finalized_indexers
+        else:
+            return grid
+
     def _validate_indexer(self, indexer) -> Union[slice, np.ndarray]:
         if isinstance(indexer, slice):
             s = indexer
@@ -723,7 +1027,64 @@ class Ugrid2d(AbstractUgrid):
 
         return indexer
 
-    def sel_points(self, x: FloatArray, y: FloatArray):
+    def _sel_box(
+        self,
+        obj,
+        x,
+        y,
+    ):
+        xmin, ymin, xmax, ymax = self.bounds
+        bounds = [
+            numeric_bound(x.start, xmin),
+            numeric_bound(y.start, ymin),
+            numeric_bound(x.stop, xmax),
+            numeric_bound(y.stop, ymax),
+        ]
+        face_index = self.locate_bounding_box(*bounds)
+        grid, indexes = self.topology_subset(face_index, return_index=True)
+        indexes = {k: v for k, v in indexes.items() if k in obj.dims}
+        new_obj = obj.isel(indexes)
+        return new_obj, grid
+
+    def _sel_yline(
+        self,
+        obj,
+        x,
+        y,
+    ):
+        xmin, _, xmax, _ = self.bounds
+        if y.size != 1:
+            raise ValueError(
+                "If x is a slice without steps, y should be a single value"
+            )
+        y = y[0]
+        xstart = numeric_bound(x.start, xmin)
+        xstop = numeric_bound(x.stop, xmax)
+        edges = np.array([[[xstart, y], [xstop, y]]])
+        _, index, xy = self.intersect_edges(edges)
+        coords, index = section_coordinates(edges, xy, self.face_dimension, index)
+        return obj.isel({self.face_dimension: index}).assign_coords(coords)
+
+    def _sel_xline(
+        self,
+        obj,
+        x,
+        y,
+    ):
+        _, ymin, _, ymax = self.bounds
+        if x.size != 1:
+            raise ValueError(
+                "If y is a slice without steps, x should be a single value"
+            )
+        x = x[0]
+        ystart = numeric_bound(y.start, ymin)
+        ystop = numeric_bound(y.stop, ymax)
+        edges = np.array([[[x, ystart], [x, ystop]]])
+        _, index, xy = self.intersect_edges(edges)
+        coords, index = section_coordinates(edges, xy, self.face_dimension, index)
+        return obj.isel({self.face_dimension: index}).assign_coords(coords)
+
+    def sel_points(self, obj, x: FloatArray, y: FloatArray):
         """
         Select points in the unstructured grid.
 
@@ -731,13 +1092,11 @@ class Ugrid2d(AbstractUgrid):
         ----------
         x: 1d array of floats with shape ``(n_points,)``
         y: 1d array of floats with shape ``(n_points,)``
+        obj: xr.DataArray or xr.Dataset
 
         Returns
         -------
-        dimension: str
-        as_ugrid: bool
-        index: 1d array of integers
-        coords: dict
+        selection: xr.DataArray or xr.Dataset
         """
         x = np.atleast_1d(x)
         y = np.atleast_1d(y)
@@ -754,9 +1113,9 @@ class Ugrid2d(AbstractUgrid):
             "x": (dim, xy[valid, 0]),
             "y": (dim, xy[valid, 1]),
         }
-        return dim, False, index, coords
+        return obj.isel({dim: index}).assign_coords(coords)
 
-    def sel(self, x=None, y=None):
+    def sel(self, obj, x=None, y=None):
         """
         Find selection in the UGRID x and y coordinates.
 
@@ -767,6 +1126,7 @@ class Ugrid2d(AbstractUgrid):
 
         Parameters
         ----------
+        obj: xr.DataArray or xr.Dataset
         x: float, 1d array, slice
         y: float, 1d array, slice
 
@@ -785,71 +1145,55 @@ class Ugrid2d(AbstractUgrid):
 
         x = self._validate_indexer(x)
         y = self._validate_indexer(y)
-        dim = self.face_dimension
-        as_ugrid = False
-        xmin, ymin, xmax, ymax = self.bounds
         if isinstance(x, slice) and isinstance(y, slice):
-            # Bounding box selection
-            # Fill None values by infinite
-            bounds = [
-                numeric_bound(x.start, xmin),
-                numeric_bound(y.start, ymin),
-                numeric_bound(x.stop, xmax),
-                numeric_bound(y.stop, ymax),
-            ]
-            as_ugrid = True
-            coords = {}
-            index = self.locate_bounding_box(*bounds)
+            f = self._sel_box
         elif isinstance(x, slice) and isinstance(y, np.ndarray):
-            if y.size != 1:
-                raise ValueError(
-                    "If x is a slice without steps, y should be a single value"
-                )
-            y = y[0]
-            xstart = numeric_bound(x.start, xmin)
-            xstop = numeric_bound(x.stop, xmax)
-            edges = np.array([[[xstart, y], [xstop, y]]])
-            _, index, xy = self.intersect_edges(edges)
-            coords, index = section_coordinates(edges, xy, dim, index)
-
+            f = self._sel_yline
         elif isinstance(x, np.ndarray) and isinstance(y, slice):
-            if x.size != 1:
-                raise ValueError(
-                    "If y is a slice without steps, x should be a single value"
-                )
-            x = x[0]
-            ystart = numeric_bound(y.start, ymin)
-            ystop = numeric_bound(y.stop, ymax)
-            edges = np.array([[[x, ystart], [x, ystop]]])
-            _, index, xy = self.intersect_edges(edges)
-            coords, index = section_coordinates(edges, xy, dim, index)
-
+            f = self._sel_xline
         elif isinstance(x, np.ndarray) and isinstance(y, np.ndarray):
             # Orthogonal points
-            yy, xx = np.meshgrid(y, x, indexing="ij")
-            return self.sel_points(xx.ravel(), yy.ravel())
-
+            y, x = [a.ravel() for a in np.meshgrid(y, x, indexing="ij")]
+            f = self.sel_points
         else:
             raise TypeError(
                 f"Invalid indexer types: {type(x).__name__}, and {type(y).__name__}"
             )
-        return dim, as_ugrid, index, coords
+        return f(obj, x, y)
 
-    def topology_subset(self, index: Union[BoolArray, IntArray]):
-        """
-        Create a new UGRID1D topology for a subset of this topology.
+    @staticmethod
+    def merge_partitions(grids):
+        from xugrid.ugrid import partitioning
 
-        Parameters
-        ----------
-        face_index: 1d array of integers or bool
-            Edges of the subset.
+        # Grab a sample grid
+        grid = next(iter(grids))
+        fill_value = grid.fill_value
+        node_coordinates, node_indexes, node_inverse = partitioning.merge_nodes(grids)
+        new_faces, face_indexes = partitioning.merge_faces(
+            grids, node_inverse, fill_value
+        )
+        indexes = {
+            grid.node_dimension: node_indexes,
+            grid.face_dimension: face_indexes,
+        }
 
-        Returns
-        -------
-        subset: Ugrid2d
-        """
-        index = np.atleast_1d(index)
-        return self._topology_subset(index, self.face_node_connectivity)
+        if grid._edge_node_connectivity is not None:
+            new_edges, edge_indexes = partitioning.merge_edges(grids, node_inverse)
+            indexes[grid.edge_dimension] = edge_indexes
+        else:
+            new_edges = None
+
+        merged_grid = Ugrid2d(
+            *node_coordinates.T,
+            fill_value,
+            new_faces,
+            name=grid.name,
+            edge_node_connectivity=new_edges,
+            projected=grid.projected,
+            crs=grid.crs,
+            attrs=grid._attrs,
+        )
+        return merged_grid, indexes
 
     def triangulate(self):
         """
@@ -927,7 +1271,7 @@ class Ugrid2d(AbstractUgrid):
 
     def refine_polygon(
         self,
-        polygon: sg.Polygon,
+        polygon: "shapely.geometry.Polygon",  # type: ignore # noqa
         min_face_size: float,
         refine_intersected: bool = True,
         use_mass_center_when_refining: bool = True,
@@ -958,7 +1302,7 @@ class Ugrid2d(AbstractUgrid):
 
     def delete_polygon(
         self,
-        polygon: sg.Polygon,
+        polygon: "shapely.geometry.Polygon",  # type: ignore # noqa
         delete_option: str = "all_face_circumenters",
         invert_deletion: bool = False,
     ):
@@ -971,7 +1315,7 @@ class Ugrid2d(AbstractUgrid):
 
     @staticmethod
     def from_polygon(
-        polygon: sg.Polygon,
+        polygon: "shapely.geometry.Polygon",  # type: ignore # noqa
     ):
         import meshkernel as mk
 
@@ -991,7 +1335,7 @@ class Ugrid2d(AbstractUgrid):
         return ugrid
 
     @staticmethod
-    def from_geodataframe(geodataframe: gpd.GeoDataFrame):
+    def from_geodataframe(geodataframe: "geopandas.GeoDataFrame"):  # type: ignore # noqa
         """
         Convert a geodataframe of polygons to UGRID2D topology.
 
@@ -1005,57 +1349,34 @@ class Ugrid2d(AbstractUgrid):
         return Ugrid2d(x, y, fill_value, face_node_connectivity)
 
     @staticmethod
-    def from_structured(data: Union[xr.DataArray, xr.Dataset]) -> "Ugrid2d":
-        """
-        Derive the 2D-UGRID quadrilateral mesh topology from a structured DataArray
-        or Dataset, with (2D-dimensions) "y" and "x".
-
-        Parameters
-        ----------
-        data: Union[xr.DataArray, xr.Dataset]
-            Structured data from which the "x" and "y" coordinate will be used to
-            define the UGRID-2D topology.
-
-        Returns
-        -------
-        ugrid_topology: Ugrid2d
-        """
-
-        def _coord(da, dim):
-            """
-            Transform N xarray midpoints into N + 1 vertex edges
-            """
-            dxs = np.diff(da[dim].values)
-            dx = dxs[0]
-            atolx = abs(1.0e-4 * dx)
-            if not np.allclose(dxs, dx, atolx):
+    def from_structured(
+        data: Union[xr.DataArray, xr.Dataset],
+        x_bounds: str = None,
+        y_bounds: str = None,
+    ) -> "Ugrid2d":
+        if x_bounds is not None and y_bounds is not None:
+            x_bounds = data[x_bounds]
+            y_bounds = data[y_bounds]
+        else:
+            x_coord, y_coord = conversion.infer_xy_coords(data)
+            x_bounds = conversion.infer_bounds(data, x_coord)
+            y_bounds = conversion.infer_bounds(data, y_coord)
+            if x_bounds is None or y_bounds is None:
                 raise ValueError(
-                    f"DataArray has to be equidistant along {dim}, or cellsizes"
-                    " must be provided as a coordinate."
+                    "Could not infer bounds. Please provide x_bounds and"
+                    " y_bounds explicitly."
                 )
-            dxs = np.full(da[dim].size, dx)
-            dxs = np.abs(dxs)
-            x = da[dim].values
-            if not da.indexes[dim].is_monotonic_increasing:
-                x = x[::-1]
-                dxs = dxs[::-1]
-            # This assumes the coordinate to be monotonic increasing
-            x0 = x[0] - 0.5 * dxs[0]
-            x = np.full(dxs.size + 1, x0)
-            x[1:] += np.cumsum(dxs)
-            return x
 
-        # Transform midpoints into vertices
-        # These are always returned monotonically increasing
-        xcoord = _coord(data, "x")
-        ycoord = _coord(data, "y")
+        nx, _ = x_bounds.shape
+        ny, _ = y_bounds.shape
+        nfaces = ny * nx
+        x = conversion.bounds_to_vertices(x_bounds)
+        y = conversion.bounds_to_vertices(y_bounds)
+
         # Compute all vertices, these are the ugrid nodes
-        node_y, node_x = (a.ravel() for a in np.meshgrid(ycoord, xcoord, indexing="ij"))
-        linear_index = np.arange(node_x.size, dtype=np.intp).reshape(
-            ycoord.size, xcoord.size
-        )
+        node_y, node_x = (a.ravel() for a in np.meshgrid(y, x, indexing="ij"))
+        linear_index = np.arange(node_x.size, dtype=np.intp).reshape((ny + 1, nx + 1))
         # Allocate face_node_connectivity
-        nfaces = (ycoord.size - 1) * (xcoord.size - 1)
         face_nodes = np.empty((nfaces, 4), dtype=IntDType)
         # Set connectivity in counterclockwise manner
         face_nodes[:, 0] = linear_index[:-1, 1:].ravel()  # upper right
