@@ -1,14 +1,85 @@
 import abc
 import copy
-from typing import Tuple, Union
+from itertools import chain
+from typing import Tuple, Type, Union
 
 import numpy as np
+import pandas as pd
 import xarray as xr
 from scipy.sparse import csr_matrix
 
-from .. import connectivity
-from ..typing import BoolArray, FloatArray, IntArray
-from . import conventions
+from xugrid.constants import BoolArray, FloatArray, IntArray
+from xugrid.ugrid import connectivity, conventions
+
+
+def as_pandas_index(index: Union[BoolArray, IntArray, pd.Index], n: int):
+    if isinstance(index, np.ndarray):
+        if index.size > n:
+            raise ValueError(
+                f"index size {index.size} is larger than dimension size: {n}"
+            )
+        if np.issubdtype(index.dtype, np.bool_):
+            # Significantly quicker if all true.
+            if index.all():
+                pd_index = pd.RangeIndex(0, n)
+            else:
+                pd_index = pd.Index((np.arange(n)[index]))
+        elif np.issubdtype(index.dtype, np.integer):
+            pd_index = pd.Index(index)
+        else:
+            raise TypeError(f"index should be bool or integer. Received: {index.dtype}")
+
+    elif isinstance(index, pd.Index):
+        pd_index = index
+
+    else:
+        raise TypeError(
+            "index should be pandas Index or numpy array. Received: "
+            f"{type(index).__name__}"
+        )
+
+    if not pd_index.is_unique:
+        raise ValueError(
+            "index contains repeated values; only subsets will result "
+            "in valid UGRID topology."
+        )
+    if not pd_index.is_monotonic_increasing:
+        raise NotImplementedError("UGRID indexes must be sorted and unique.")
+
+    return pd_index
+
+
+def align(obj, grids, old_indexes):
+    """
+    Check which indexes have changed. Index on those new values.
+    If none are changed, return (obj, grids) as is.
+    """
+    if old_indexes is None:
+        return obj, grids
+
+    ugrid_dims = set(
+        chain.from_iterable(grid.dimensions for grid in grids)
+    ).intersection(old_indexes)
+    new_indexes = {
+        k: index
+        for k, index in obj.indexes.items()
+        if (k in ugrid_dims) and (not index.equals(old_indexes[k]))
+    }
+    if not new_indexes:
+        return obj, grids
+
+    # Group the indexers by grid
+    new_grids = []
+    for grid in grids:
+        ugrid_dims = set(grid.dimensions).intersection(new_indexes)
+        ugrid_indexes = {dim: new_indexes[dim] for dim in ugrid_dims}
+        newgrid, indexers = grid.isel(indexers=ugrid_indexes, return_index=True)
+        indexers = {
+            k: v for k, v in indexers.items() if k in obj.dims and k not in new_indexes
+        }
+        obj = obj.isel(indexers)
+        new_grids.append(newgrid)
+    return obj, new_grids
 
 
 class AbstractUgrid(abc.ABC):
@@ -16,6 +87,10 @@ class AbstractUgrid(abc.ABC):
     def topology_dimension():
         """ """
 
+    @abc.abstractproperty
+    def core_dimension():
+        """ """
+
     @abc.abstractproperty
     def dimensions():
         """ """
@@ -29,21 +104,49 @@ class AbstractUgrid(abc.ABC):
         """ """
 
     @abc.abstractmethod
-    def isel():
+    def topology_subset():
+        """ """
+
+    @abc.abstractmethod
+    def clip_box():
+        """ """
+
+    @abc.abstractmethod
+    def sel_points():
         """ """
 
     @abc.abstractmethod
     def sel():
         """ """
 
-    @abc.abstractmethod
-    def topology_subset():
-        """ """
-
     @abc.abstractmethod
     def _clear_geometry_properties():
         """ """
 
+    @abc.abstractstaticmethod
+    def merge_partitions():
+        """ """
+
+    def _initialize_indexes_attrs(self, name, dataset, indexes, attrs):
+        defaults = conventions.default_topology_attrs(name, self.topology_dimension)
+        if dataset is None:
+            if attrs is not None:
+                defaults.update(attrs)
+            x, y = defaults["node_coordinates"].split(" ")
+            self._indexes = {"node_x": x, "node_y": y}
+            self._attrs = defaults
+        else:
+            if attrs is not None:
+                raise ValueError("Provide either dataset or attrs, not both.")
+            if indexes is None:
+                raise ValueError("indexes must be provided for dataset")
+            derived_dims = dataset.ugrid_roles.dimensions[name]
+            self._indexes = indexes
+            self._attrs = {**defaults, **derived_dims, **dataset[name].attrs}
+        # Ensure the name is always in sync.
+        self._attrs["name"] = name
+        return
+
     @staticmethod
     def _single_topology(dataset: xr.Dataset):
         topologies = dataset.ugrid_roles.topology
@@ -210,26 +313,24 @@ class AbstractUgrid(abc.ABC):
             raise ValueError("connectivity contains negative values")
         return da.copy(data=cast)
 
-    def _topology_subset(
-        self, indices: Union[BoolArray, IntArray], node_connectivity: IntArray
-    ):
-        is_same = False
-        if np.issubdtype(indices.dtype, np.bool_):
-            is_same = indices.all()
-        else:
-            # TODO: check for unique indices if integer?
-            is_same = np.array_equal(indices, np.arange(node_connectivity.shape[0]))
+    def _precheck(self, multi_index):
+        dim, index = multi_index.popitem()
+        for check_dim, check_index in multi_index.items():
+            if not index.equals(check_index):
+                raise ValueError(
+                    f"UGRID dimensions do not align: {dim} versus {check_dim}"
+                )
+        return index
 
-        if is_same:
-            return self
-        # Subset of faces, create new topology data
-        else:
-            subset = node_connectivity[indices]
-            node_indices = np.unique(subset.ravel())
-            new_connectivity = connectivity.renumber(subset)
-            node_x = self.node_x[node_indices]
-            node_y = self.node_y[node_indices]
-            return self.__class__(node_x, node_y, self.fill_value, new_connectivity)
+    def _postcheck(self, indexers, finalized_indexers):
+        for dim, indexer in indexers.items():
+            if dim != self.core_dimension:
+                if not indexer.equals(finalized_indexers[dim]):
+                    raise ValueError(
+                        f"This subset selection of UGRID dimension {dim} results"
+                        "in an invalid topology "
+                    )
+        return
 
     def set_node_coords(
         self,
@@ -475,6 +576,9 @@ class AbstractUgrid(abc.ABC):
         **kwargs : optional
             Additional keyword arguments to ``matplotlib.pyplot.line``.
         """
-        from ..plot import line
+        from xugrid.plot import line
 
         return line(self, **kwargs)
+
+
+UgridType = Type[AbstractUgrid]
